/* Copyright (C) 1998-9 Pancrazio `Ezio' de Mauro <p@demauro.net>
 * Copyright (C) 2004 Milosz `Dem' Tanski <mtanski@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: helper.c,v 1.3 2004/12/20 02:33:45 demie Exp $
 *
 * -----------------------------------------------------------------
 * helper.c: Helper functions
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <fcntl.h>
#include <dlfcn.h>

#include "zengarden.h"

void logger(const char *format, ...) {
  char *msg;
  va_list ap;
  FILE *log;

  va_start(ap, format);
  if (vasprintf(&msg, format, ap) < 0) {
    syslog(LOGLEVEL, "%i: Could not write `%s' in `%s': %s\n", (__LINE__-1),msg, logfile, strerror(errno));
    _exit(EXIT_FAILURE);
  }
  va_end(ap);

  if (logfile)
    if((log = true_fopen(logfile, "a"))) {
      if(fputs(msg, log) == EOF)
	syslog(LOGLEVEL, "%i: Could not write `%s' in `%s': %s\n", (__LINE__-1),msg, logfile, strerror(errno));
      if(fclose(log))
        syslog(LOGLEVEL, "Could not close `%s': %s\n", logfile, strerror(errno));
    } else
      syslog(LOGLEVEL, "Could not open `%s' to write `%s': %s\n", logfile, msg, strerror(errno));

  free(msg);
}

/* free this bad boy */
char *canonicalize(const char *path) {
  char *canonic = 0;
  char *cwd = 0;
  char *can = canonicalize_file_name(path);
  
  CHECKINIT;

  if (!can && (path[0] != '/')) {
    /* The path could not be canonicalized, append it
     * to the current working directory if it was not 
     * an absolute path */
     if (!(cwd = getcwd(NULL, 0))) {
       syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
       _exit(EXIT_FAILURE);
     }

     if (asprintf(&canonic ,"%s/%s", cwd, path) < 0) {
       syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
       _exit(EXIT_FAILURE);
     }

     free(cwd);
  } else if (can) {
    canonic = can;
  } else {
    canonic = strdup(path);
  }
  
  return canonic;
} 

/* The function calling this should free returned char * */
char *chpath(const char *prefix,const char *postfix)
{
  char *tmp, *end;
  int len = strlen(postfix);
  if (!(tmp = malloc(irlen + 1 + len + 1))) {
    syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
    _exit(EXIT_FAILURE);
  }
  
  end = mempcpy(tmp, prefix, sizeof(char) * irlen);
  end = mempcpy(end, (char *) '/', sizeof(char));
  end = mempcpy(end, postfix, sizeof(char) * len);
  end = mempcpy(end, (char *) EOF, sizeof(char));

  return tmp;
}

void make_path (const char *path) {
  char *dir = strdup(path);

  if (!dir) {
    syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
    _exit(EXIT_FAILURE);
  }

  for (int i=0; path[i]; i++) {
    if (path[i] == '/') {
      dir[i+1] = 0;                /* we found new depth */
      if(access(dir, F_OK))        /* exists ?*/
        true_mkdir(dir, S_IRWXU);
      dir[i+1] = path[i+1];        /* silly but faster than using strdup everytime */
    }
  }
  
  free(dir);
}

void backup(const char *path) {
  struct stat in_inode;
  char *backup_path, *dir, *tmp;
  void *from;
  int in, out;
  
  /* Are be backing up? -- Can't backup when doing chroot */
  if (!bpath || instroot)
    return;

  if ((strstr (path, "/dev") == path)     /* Check if this is inside /dev */
       || (strstr (path, "/tmp") == path) /* Now check for /tmp */
       || (strstr (path, bpath))) {    /* Finally, the backup path itself */
      return; 
  }

  /* copy a string we'll need later */
  if (!(dir = strdup(path))) {
    syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
    _exit(EXIT_FAILURE);
  }

  // TODO: find a better solution, than using a no-backup directory
  if (asprintf(&backup_path, "%s/no-backup/%s", bpath, path) < 0) {
    syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
    _exit(EXIT_FAILURE);
  }
  
//TODO: rewrite this shit
  /* Does it exist already? */
  if (access(path, F_OK)) {
    /* It doesn't exist, we'll tag it so we won't back it up  */
    /* if we run into it later                                */
    make_path(backup_path);

    /* wish glibc had touch function */
    int p = true_creat(backup_path, S_IREAD);
    if (!(p < 0)) close (p);
   
    return;
  }

  /* Is this one tagged for no backup (i.e. it didn't previously exist)? */
  if (access(backup_path, F_OK) == 0) {
    return;
  }
//TODO: end


  free(backup_path); /* re use */

  /* Append the path to the backup_path */
  if (asprintf(&backup_path, "%s%s", bpath, path) < 0) {
    syslog(LOGLEVEL, "%i: Could not allocate enouth memory: %s\n", (__LINE__-1), strerror(errno));
    _exit(EXIT_FAILURE);
  }

  /* Create the directory tree for this file in the backup dir */
  make_path(backup_path);
  
  stat(path, &in_inode);
  switch (in_inode.st_mode & S_IFMT) {
    /* regular file */
    case S_IFREG:
      /* input file */
      if ((in = true_open (path, O_RDONLY)) < 0) {
        syslog(LOGLEVEL, "%i: Could not open `%s' : %s\n", (__LINE__-1), path, strerror(errno));
	
	free(backup_path);
	free(dir);
	
        return;
      } else if ((from = mmap(NULL, in_inode.st_size, PROT_READ, MAP_PRIVATE, in, 0)) == NULL) {
        syslog(LOGLEVEL,"%i: Could not mmap file %s : %s\n", (__LINE__-1), path, strerror(errno));
	
	free(backup_path);
	free(dir);
        close (in);

	return;
      }
      
      /* dest file */
      if ((out = true_open (backup_path, O_WRONLY|O_CREAT|O_TRUNC, in_inode.st_mode)) < 0) {
        syslog(LOGLEVEL, "%i: Could not open `%s' : %s\n", (__LINE__-1), backup_path, strerror(errno));
	
	free(backup_path);
	free(dir);
        munmap(from, in_inode.st_size);
        close(in);
	
	return;
      } 
     
      /* a couple hints to speed things up */
      madvise(from, in_inode.st_size, POSIX_MADV_SEQUENTIAL);
    
      /* write the data */
      if (write(out, from, in_inode.st_size) != in_inode.st_size) {
        syslog(LOGLEVEL, "%i: Could write to `%s' : %s\n", (__LINE__-1), backup_path, strerror(errno));
      
        munmap(from, in_inode.st_size);
        close(in);
        close(out);
        free(backup_path);
	free(dir);
	
        return;
      }
     
      /* cleanup */
      munmap(from, in_inode.st_size);
      close(in);
      close(out);

      break;
    /* directory */
    case S_IFDIR:
      true_mkdir(backup_path, in_inode.st_mode);
      break;
    /* block device */
    case S_IFBLK:
      mknod(backup_path, in_inode.st_mode | S_IFBLK, in_inode.st_rdev);
      break;
    /* char device */
    case S_IFCHR:
      mknod(backup_path, in_inode.st_mode | S_IFCHR, in_inode.st_rdev);
      break;
    /* pipe */
    case S_IFIFO:
      mknod(backup_path, in_inode.st_mode | S_IFIFO, 0);
  }

  /* If we have a new file, set it to the right owner and group */
  if (!access(path, F_OK))
    true_chown(backup_path, in_inode.st_uid, in_inode.st_gid);

  /* Check the owner and permission of the created directories */
  for (int i=0 ; path[i]; i++)
    if (path[i] == '/') {
      dir[i+1] = 0;
      if (!stat(dir, &in_inode)) {
        backup_path[bplen+i+1] = 0;
        true_chmod (backup_path, in_inode.st_mode);
	true_chown (backup_path, in_inode.st_uid, in_inode.st_gid);
	backup_path[bplen+i+1] = path[i+1]; /* silly sol. */
      }
      dir[i+1] = path[i+1]; /* silly sol. */
    }
 
  free(dir);
  free(backup_path);
}

void check_jail(const char *path, const char *func) {
  int fail = 1;
  char *tmp = canonicalize(path);

  if (!bjail)
    return;

  if (!strncmp(bjail, tmp, bjlen))
    fail = 0;

  /* allow using of /tmp/ - last / important
   * /dev/null
   * /dev/tty
   */
  if (!strncmp("/tmp/", tmp, 5))
    fail = 0;
  if (!strncmp("/dev/null", tmp, 9))
    fail = 0;
  if (!strncmp("/dev/tty", tmp, 8))
    fail = 0;
 
  if (fail) {
    printf("Error: %s outside the ZG_BUILD_PATH: %s %s\n", func, tmp, bjail);
    logger("die: %s outside the ZG_BUILD_PATH: %s\n", func, tmp);
    _exit(EXIT_FAILURE);
  }

  free(tmp);
}

// vim:tabstop=2:shiftwidth=2:expandtab:
